import torch
import torch.nn as nn

class InputTransition(nn.Module):
    def __init__(
            self, in_channels, out_channels, act, bias=False
    ):
        super(InputTransition, self).__init__()
        if out_channels % in_channels != 0:
            raise ValueError(
                f"out channels should be divisible by in_channels. Got in_channels={in_channels}, out_channels={out_channels}."
            )
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.conv1 = nn.Conv3d(in_channels, out_channels, kernel_size=5, padding=2, bias=bias)
        self.bn1 = nn.BatchNorm3d(out_channels)
        self.act1 = act

    def forward(self, x):
        out = self.bn1(self.conv1(x))
        repeat_num = self.out_channels // self.in_channels
        x16 = x.repeat([1, repeat_num, 1, 1, 1][: 5])
        out = self.act1(torch.add(out, x16))
        return out

class LUConv(nn.Module):
    def __init__(self, nchan, bias=False):
        super().__init__()
        self.conv1 = nn.Conv3d(nchan, nchan, kernel_size=5, padding=2, bias=bias)
        self.bn1 = nn.BatchNorm3d(nchan)
        self.act = nn.ELU(inplace=True)

    def forward(self, x):
        return self.act(self.bn1(self.conv1(x)))

def _make_nconv(nchan, depth, bias=False):
    layers = []
    for _ in range(depth):
        layers.append(LUConv(nchan, bias))
    return nn.Sequential(*layers)

class DownTransition(nn.Module):
    def __init__(
            self,
            in_channels,
            nconvs,
            act,
            dropout=None,
            bias=False,
    ):
        super().__init__()

        out_channels = 2 * in_channels
        self.down_conv = nn.Conv3d(in_channels, out_channels, kernel_size=2, stride=2, bias=bias)
        self.bn1 = nn.BatchNorm3d(out_channels)
        self.act = act
        self.dropout = dropout
        self.do1 = nn.Dropout3d()
        self.ops = _make_nconv(out_channels, nconvs, bias)

    def forward(self, x):
        down = self.act(self.bn1(self.down_conv(x)))
        if self.dropout:
            down = self.do1(down)
        out = self.ops(down)
        out = self.act(torch.add(out, down))
        return out

class UpTransition(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            nconvs,
            act,
            dropout=None,
    ):
        super().__init__()
        self.up_conv = nn.ConvTranspose3d(in_channels, out_channels // 2, kernel_size=2, stride=2)
        self.norm = nn.BatchNorm3d(out_channels // 2)
        self.dropout_type = dropout
        self.dropout = nn.Dropout3d()
        self.act = act
        self.ops = _make_nconv(out_channels, nconvs)
        self.do1 = nn.Dropout()
        self.do2 = nn.Dropout3d()


    def forward(self, x, skipx):
        if self.dropout_type:
            x = self.do1(x)
        skipxdo = self.do2(skipx)
        out = self.act(self.norm(self.up_conv(x)))
        xcat = torch.cat((out, skipxdo), 1)
        out = self.ops(xcat)
        out = self.act(torch.add(out, xcat))
        return out


class OutputTransition(nn.Module):
    def __init__(
            self, in_channels, out_channels, act, bias=False
    ):
        super().__init__()
        self.conv1 = nn.Conv3d(in_channels, out_channels, kernel_size=5, padding=2, bias=bias)
        self.bn1 = nn.BatchNorm3d(out_channels)
        self.conv2 = nn.Conv3d(out_channels, out_channels, kernel_size=1)
        self.act = act

    def forward(self, x):
        # convolve 32 down to 2 channels
        out = self.act(self.bn1(self.conv1(x)))
        out = self.conv2(out)
        return out


class VNet3D(nn.Module):
    """
    V-Net based on `Fully Convolutional Neural Networks for Volumetric Medical Image Segmentation
    <https://arxiv.org/pdf/1606.04797.pdf>`.
    """
    def __init__(
            self,
            in_channels=1,
            num_classes=1,
            act=nn.ELU(inplace=True),
            dropout=True,
            bias=False,
    ):
        super().__init__()
        self.relu = nn.ReLU(inplace=True)
        self.in_tr = InputTransition(in_channels, 16, self.relu, bias=bias)
        self.down_tr32 = DownTransition(16, 1, self.relu, bias=bias)
        self.down_tr64 = DownTransition(32, 2, self.relu, bias=bias)
        self.down_tr128 = DownTransition(64, 3, act, dropout=dropout, bias=bias)
        self.down_tr256 = DownTransition(128, 2, act, dropout=dropout, bias=bias)
        self.up_tr256 = UpTransition(256, 256, 2, act, dropout=dropout)
        self.up_tr128 = UpTransition(256, 128, 2, act, dropout=dropout)
        self.up_tr64 = UpTransition(128, 64, 1, self.relu)
        self.up_tr32 = UpTransition(64, 32, 1, self.relu)
        self.out_tr = OutputTransition(32, num_classes, act, bias=bias)

    def forward(self, x):
        out16 = self.in_tr(x)
        out32 = self.down_tr32(out16)
        out64 = self.down_tr64(out32)
        out128 = self.down_tr128(out64)
        out256 = self.down_tr256(out128)
        x = self.up_tr256(out256, out128)
        x = self.up_tr128(x, out64)
        x = self.up_tr64(x, out32)
        x = self.up_tr32(x, out16)
        x = self.out_tr(x)
        return x


if __name__ == '__main__':
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    # B x C x Z x Y x X
    # 4 x 1 x 64 x 64 x 64
    input = torch.ones(1, 1, 32, 32, 32).to(device)
    net = VNet3D(in_channels=1, num_classes=3)
    net = net.to(device)
    out = net(input)
    print(out.shape)
    torchsummary.summary(net, input_size=(1, 32, 32, 32))



"""
MLP module w/ dropout and configurable activation layer
"""
import torch
import torch.nn as nn
from pyzjr.core import to_2tuple
from functools import partial

class Mlp(nn.Module):
    """
    MLP as used in Vision Transformer, MLP-Mixer and related networks
    """
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class GluMlp(nn.Module):
    """
    MLP w/ GLU style gating
    See: https://arxiv.org/abs/1612.08083, https://arxiv.org/abs/2002.05202
    """
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.Sigmoid, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        assert hidden_features % 2 == 0
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features // 2, out_features)
        self.drop = nn.Dropout(drop)
        self.init_weights()

    def init_weights(self):
        # override init of fc1 w/ gate portion set to weight near zero, bias=1
        fc1_mid = self.fc1.bias.shape[0] // 2
        nn.init.ones_(self.fc1.bias[fc1_mid:])
        nn.init.normal_(self.fc1.weight[fc1_mid:], std=1e-6)

    def forward(self, x):
        x = self.fc1(x)
        x, gates = x.chunk(2, dim=-1)
        x = x * self.act(gates)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class GatedMlp(nn.Module):
    """MLP as used in gMLP"""
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU,
                 gate_layer=None, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        if gate_layer is not None:
            assert hidden_features % 2 == 0
            self.gate = gate_layer(hidden_features)
            hidden_features = hidden_features // 2  # FIXME base reduction on gate property?
        else:
            self.gate = nn.Identity()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.gate(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class ConvMlp(nn.Module):
    """
    MLP using 1x1 convs that keeps spatial dims
    """
    def __init__(
            self, in_features, hidden_features=None, out_features=None, act_layer=nn.ReLU, norm_layer=None, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=1, bias=True)
        self.norm = norm_layer(hidden_features) if norm_layer else nn.Identity()
        self.act = act_layer()
        self.fc2 = nn.Conv2d(hidden_features, out_features, kernel_size=1, bias=True)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        return x

class ConvMod(nn.Module):
    """come from pyzjr.backbone.conv2former"""
    def __init__(self, dim):
        super().__init__()

        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.a = nn.Sequential(
            nn.Conv2d(dim, dim, 1),
            nn.GELU(),
            nn.Conv2d(dim, dim, 11, padding=5, groups=dim)
        )
        self.v = nn.Conv2d(dim, dim, 1)
        self.proj = nn.Conv2d(dim, dim, 1)

    def forward(self, x):
        x = self.norm(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        a = self.a(x)
        x = a * self.v(x)
        x = self.proj(x)
        return x


class GlobalResponseNormMlp(nn.Module):
    """ MLP w/ Global Response Norm (see grn.py), nn.Linear or 1x1 Conv2d
    """
    def __init__(
            self,
            in_features,
            hidden_features=None,
            out_features=None,
            act_layer=nn.GELU,
            bias=True,
            drop=0.,
            use_conv=False,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)
        drop_probs = to_2tuple(drop)
        linear_layer = partial(nn.Conv2d, kernel_size=1) if use_conv else nn.Linear

        self.fc1 = linear_layer(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])
        self.grn = GlobalResponseNorm(hidden_features, channels_last=not use_conv)
        self.fc2 = linear_layer(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.grn(x)
        x = self.fc2(x)
        x = self.drop2(x)
        return x


class GlobalResponseNorm(nn.Module):
    """ Global Response Normalization layer
    """
    def __init__(self, dim, eps=1e-6, channels_last=True):
        super().__init__()
        self.eps = eps
        if channels_last:
            self.spatial_dim = (1, 2)
            self.channel_dim = -1
            self.wb_shape = (1, 1, 1, -1)
        else:
            self.spatial_dim = (2, 3)
            self.channel_dim = 1
            self.wb_shape = (1, -1, 1, 1)

        self.weight = nn.Parameter(torch.zeros(dim))
        self.bias = nn.Parameter(torch.zeros(dim))

    def forward(self, x):
        x_g = x.norm(p=2, dim=self.spatial_dim, keepdim=True)
        x_n = x_g / (x_g.mean(dim=self.channel_dim, keepdim=True) + self.eps)
        return x + torch.addcmul(self.bias.view(self.wb_shape), self.weight.view(self.wb_shape), x * x_n)


import torch
import os
import numpy as np
from torch.utils.data import Dataset, DataLoader
from pyzjr.torch3d.nii.nii_utils import nii_load, center_scale_nii, center_crop_nii
from pyzjr.nn.torchutils.OneHot import one_hot_3d

class NiftiDataset3D(Dataset):
    """
    用于加载3D Nifti格式图像数据的PyTorch数据集类
    默认是采用了中心裁剪和中心缩放的方式来减少内存需求，所以，
    ‘ target_shape ’ = 裁剪大小 * 缩放比例， 一般请勿修改缩放比例

    Args:
        image_path (str): 包含图像文件的目录路径。
        label_path (str): 包含标签文件的目录路径。
        target_shape (tuple): 所需的图像和标签的目标形状。
        scale (float): 缩放因子，用于缩放图像和标签。默认为0.5.
        num_classes (int): 类别数量。
        normalize (bool): 是否对图像进行归一化。默认为True.
    """
    def __init__(self,
                 image_path,
                 label_path,
                 target_shape,
                 scale=.5,
                 num_classes=None,
                 normalize=True
                 ):
        super().__init__()
        self.image_paths = self.read_path2list(image_path)
        self.label_paths = self.read_path2list(label_path)
        self.num_classes = num_classes
        self.normalize = normalize
        self.scale = scale
        self.target_crop_shape = (int(target_shape[0] // scale),
                                 int(target_shape[1] // scale),
                                 int(target_shape[2] // scale))

    def __getitem__(self, item):
        one_image, one_label = self.read_nii_im_and_label(
            self.image_paths[item], self.label_paths[item]
        )
        one_image = center_crop_nii(one_image, self.target_crop_shape)
        one_label = center_crop_nii(one_label, self.target_crop_shape)
        one_image = center_scale_nii(one_image, self.scale)
        one_label = center_scale_nii(one_label, self.scale)
        # z, y, x = one_image.shape
        if self.normalize:
            one_image = self.normalize_image(one_image)

        image_trans = one_image[np.newaxis, :, :, :]  # b c w w h
        self.label_trans = one_label[np.newaxis, :, :, :]  # b c d w h

        if isinstance(image_trans, torch.Tensor):
            image_trans = image_trans.numpy()
        if isinstance(self.label_trans, torch.Tensor):
            self.label_trans = self.label_trans.numpy()
        one_hot_label_trans = one_hot_3d(self.label_trans[0], num_classes=self.num_classes)
        # one_hot_label_trans  b  d  y  x  c ----->  b  c  d  y  x
        return image_trans, np.transpose(one_hot_label_trans, (3, 0, 1, 2))

    def read_nii_im_and_label(self, image_path, label_path):
        image, _, _ = nii_load(image_path)
        label, _, _ = nii_load(label_path)
        # w, h, d  ----->  d, h, w
        return np.transpose(image, (2, 1, 0)), np.transpose(label, (2, 1, 0))

    def __len__(self):
        return len(self.image_paths)

    def read_path2list(self, path):
        image_files = sorted(os.listdir(path))
        return [os.path.join(path, f) for f in image_files]

    def normalize_image(self, image, clip_range=(0, 1)):
        image = image.astype(np.float32)
        image /= 255
        image = np.clip(image, clip_range[0], clip_range[1])
        return image

    def _get_label_unique(self):
        return np.unique(self.label_trans)

if __name__=="__main__":
    image_path = r"D:\PythonProject\pytorch_segmentation_3D\DSCNet_3D\SegDataset\MRI_Hippocampus\imagesTe"
    label_path = r"D:\PythonProject\pytorch_segmentation_3D\DSCNet_3D\SegDataset\MRI_Hippocampus\labelsTe"

    dataset = NiftiDataset3D(image_path, label_path, target_shape=(64, 64, 64), num_classes=3)
    dataloader = DataLoader(dataset, batch_size=2, shuffle=True)

    for i, (image, onehotlabel) in enumerate(dataset):
        print(f"Sample {i+1}:")
        print("Image shape:", image.shape)
        print("One Hot Label shape:", onehotlabel.shape)
        print(np.unique(onehotlabel), np.unique(onehotlabel[0]), np.unique(onehotlabel[1]),
              np.unique(onehotlabel[2]))
        print(dataset._get_label_unique())
        print("-------------------")



"""
Visualizing and Understanding Convolutional Networks
    https://arxiv.org/pdf/1311.2901.pdf
"""
import torch
import torch.nn as nn

class ZFNet(nn.Module):
    """
    ZFNet网络结构和AlexNet保持一致，但是卷积核的大小和步长发生了变化。
    主要改进：
        ➢ 将第一个卷积层的卷积核大小改为了7×7
        ➢ 将第二、第三个卷积层的卷积步长都设置为2
        ➢ 增加了第三、第四个卷积层的卷积核个数
    """
    def __init__(self, num_classes=1000):
        super(ZFNet, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(3, 48, kernel_size=7, stride=2, padding=1),   # [-1, 48, 110, 110]
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),       # [-1, 48, 55, 55]
            nn.Conv2d(48, 128, kernel_size=5, stride=2),            # [-1, 128, 26, 26]
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),       # [-1, 128, 13, 13]
            nn.Conv2d(128, 192, kernel_size=3, padding=1),          # [-1, 192, 13, 13]
            nn.ReLU(inplace=True),
            nn.Conv2d(192, 192, kernel_size=3, padding=1),          # [-1, 192, 13, 13]
            nn.ReLU(inplace=True),
            nn.Conv2d(192, 128, kernel_size=3, padding=1),          # [-1, 128, 13, 13]
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),                  # [-1, 128, 6, 6]
        )
        self.classifier = nn.Sequential(
            nn.Dropout(p=0.5),
            nn.Linear(128 * 6 * 6, 2048),
            nn.ReLU(inplace=True),
            nn.Dropout(p=0.5),
            nn.Linear(2048, 2048),
            nn.ReLU(inplace=True),
            nn.Linear(2048, num_classes),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')  # 何教授方法
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

if __name__=='__main__':
    import torchsummary
    input = torch.ones(2, 3, 224, 224).cpu()
    net = ZFNet(num_classes=4)
    net = net.cpu()
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Total params: 14,579,268


import numpy as np
import torch
from torch import nn
from torch.nn import init



class Depth_Pointwise_Conv1d(nn.Module):
    def __init__(self,in_ch,out_ch,k):
        super().__init__()
        if(k==1):
            self.depth_conv=nn.Identity()
        else:
            self.depth_conv=nn.Conv1d(
                in_channels=in_ch,
                out_channels=in_ch,
                kernel_size=k,
                groups=in_ch,
                padding=k//2
            )
        self.pointwise_conv=nn.Conv1d(
            in_channels=in_ch,
            out_channels=out_ch,
            kernel_size=1,
            groups=1
        )
    def forward(self,x):
        out=self.pointwise_conv(self.depth_conv(x))
        return out



class MUSEAttention(nn.Module):

    def __init__(self, d_model, d_k, d_v, h,dropout=.1):


        super(MUSEAttention, self).__init__()
        self.fc_q = nn.Linear(d_model, h * d_k)
        self.fc_k = nn.Linear(d_model, h * d_k)
        self.fc_v = nn.Linear(d_model, h * d_v)
        self.fc_o = nn.Linear(h * d_v, d_model)
        self.dropout=nn.Dropout(dropout)

        self.conv1=Depth_Pointwise_Conv1d(h * d_v, d_model,1)
        self.conv3=Depth_Pointwise_Conv1d(h * d_v, d_model,3)
        self.conv5=Depth_Pointwise_Conv1d(h * d_v, d_model,5)
        self.dy_paras=nn.Parameter(torch.ones(3))
        self.softmax=nn.Softmax(-1)

        self.d_model = d_model
        self.d_k = d_k
        self.d_v = d_v
        self.h = h

        self.init_weights()


    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, queries, keys, values, attention_mask=None, attention_weights=None):

        #Self Attention
        b_s, nq = queries.shape[:2]
        nk = keys.shape[1]

        q = self.fc_q(queries).view(b_s, nq, self.h, self.d_k).permute(0, 2, 1, 3)  # (b_s, h, nq, d_k)
        k = self.fc_k(keys).view(b_s, nk, self.h, self.d_k).permute(0, 2, 3, 1)  # (b_s, h, d_k, nk)
        v = self.fc_v(values).view(b_s, nk, self.h, self.d_v).permute(0, 2, 1, 3)  # (b_s, h, nk, d_v)

        att = torch.matmul(q, k) / np.sqrt(self.d_k)  # (b_s, h, nq, nk)
        if attention_weights is not None:
            att = att * attention_weights
        if attention_mask is not None:
            att = att.masked_fill(attention_mask, -np.inf)
        att = torch.softmax(att, -1)
        att=self.dropout(att)

        out = torch.matmul(att, v).permute(0, 2, 1, 3).contiguous().view(b_s, nq, self.h * self.d_v)  # (b_s, nq, h*d_v)
        out = self.fc_o(out)  # (b_s, nq, d_model)

        v2=v.permute(0,1,3,2).contiguous().view(b_s,-1,nk) #bs,dim,n
        self.dy_paras=nn.Parameter(self.softmax(self.dy_paras))
        out2=self.dy_paras[0]*self.conv1(v2)+self.dy_paras[1]*self.conv3(v2)+self.dy_paras[2]*self.conv5(v2)
        out2=out2.permute(0,2,1) #bs.n.dim

        out=out+out2
        return out


if __name__ == '__main__':
    input=torch.randn(50,49,512)
    sa = MUSEAttention(d_model=512, d_k=512, d_v=512, h=8)
    output=sa(input,input,input)
    print(output.shape)




"""
Original paper addresshttps: <https://arxiv.org/pdf/2108.02456.pdf>

"""
import torch
from torch import nn

class ResidualAttention(nn.Module):
    def __init__(self, channel=512, num_class=1000, la=0.2):
        super().__init__()
        self.la = la
        self.fc = nn.Conv2d(in_channels=channel,
                          out_channels=num_class,
                          kernel_size=1,
                          stride=1,
                          bias=False)

    def forward(self, x):
        y_raw = self.fc(x).flatten(2) #b,num_class,hxw
        y_avg = torch.mean(y_raw, dim=2) #b,num_class
        y_max = torch.max(y_raw, dim=2)[0] #b,num_class
        score = y_avg+self.la*y_max
        return score

if __name__ == '__main__':
    input = torch.randn(50,512,7,7)
    resatt = ResidualAttention(channel=512, num_class=1000, la=0.2)
    output = resatt(input)
    print(output.shape)


import torch
import torch.nn as nn
import numpy as np
from pyzjr.core.general import is_numpy, is_tensor
from pyzjr.nn.Metrics.indexutils import ignore_background

def get_confusion_matrix_3d_torch(y_pred, y, include_background=True):
    """
    Calculate the three-dimensional confusion matrix of tensor type.
    """
    if not include_background:
        y_pred, y = ignore_background(y_pred=y_pred, y=y)

    y = y.float()
    y_pred = y_pred.float()

    if y.shape != y_pred.shape:
        raise ValueError(f"y_pred and y should have same shapes, got {y_pred.shape} and {y.shape}.")

    batch_size, n_class = y_pred.shape[:2]
    y_pred = y_pred.reshape(batch_size, n_class, -1)
    y = y.reshape(batch_size, n_class, -1)
    tp = ((y_pred + y) == 2).float()
    tn = ((y_pred + y) == 0).float()

    tp = tp.sum(dim=[2])
    tn = tn.sum(dim=[2])
    p = y.sum(dim=[2])
    n = y.shape[-1] - p

    fn = p - tp
    fp = n - tn

    return torch.stack([tp, fp, tn, fn], dim=-1)

def get_confusion_matrix_3d_np(y_pred, y, include_background=True):
    """
    Calculate the three-dimensional confusion matrix of ndarray type.
    """
    if not include_background:
        y_pred, y = ignore_background(y_pred=y_pred, y=y)

    y = y.astype(np.float32)
    y_pred = y_pred.astype(np.float32)

    if y.shape != y_pred.shape:
        raise ValueError(f"y_pred and y should have same shapes, got {y_pred.shape} and {y.shape}.")

    batch_size, n_class = y_pred.shape[:2]
    y_pred = y_pred.reshape(batch_size, n_class, -1)
    y = y.reshape(batch_size, n_class, -1)
    tp = ((y_pred + y) == 2).astype(np.float32)
    tn = ((y_pred + y) == 0).astype(np.float32)

    tp = np.sum(tp, axis=2)
    tn = np.sum(tn, axis=2)
    p = np.sum(y, axis=2)
    n = y.shape[-1] - p

    fn = p - tp
    fp = n - tn

    return np.stack([tp, fp, tn, fn], axis=-1)

def get_confusion_matrix_3d(y_pred, y, include_background=True):
    """
    Calculate the three-dimensional confusion matrix.

    Args:
    y_pred: input data to compute. It must be one-hot format and first dim is batch.
        The values should be binarized.
    y: ground truth to compute the metric. It must be one-hot format and first dim is batch.
        The values should be binarized.
    include_background: whether to include metric computation on the first channel of
        the predicted output. Defaults to True.
    """
    if is_numpy(y_pred) and is_numpy(y):
        return get_confusion_matrix_3d_np(y_pred, y, include_background)
    if is_tensor(y_pred) and is_tensor(y):
        return get_confusion_matrix_3d_torch(y_pred, y, include_background)

class ConfusionMatrixs3D(nn.Module):
    """
    This function is used to compute confusion matrix related metric.

    Args:
        metric_name: [``"sensitivity"``, ``"specificity"``, ``"precision"``, ``"negative predictive value"``,
        ``"miss rate"``, ``"fall out"``, ``"false discovery rate"``, ``"false omission rate"``,
        ``"prevalence threshold"``, ``"threat score"``, ``"accuracy"``, ``"balanced accuracy"``,
        ``"f1 score"``, ``"matthews correlation coefficient"``, ``"fowlkes mallows index"``,
        ``"informedness"``, ``"markedness"``]

    For more names, you can refer to the "_check_confusion_matrix_metric name" function
    under this category.
    """
    def __init__(self, metric_name, reduction='mean', eps=1e-5):
        super(ConfusionMatrixs3D, self).__init__()
        self.metric = self._check_confusion_matrix_metric_name(metric_name)
        self.reduction = reduction
        self.eps = eps

    def forward(self, y_pred, y):
        confusion_matrix = get_confusion_matrix_3d_torch(y_pred, y)
        confusion_matrix = self._check_conf_matrix_format(confusion_matrix)
        tp = confusion_matrix[..., 0]
        fp = confusion_matrix[..., 1]
        tn = confusion_matrix[..., 2]
        fn = confusion_matrix[..., 3]
        p = tp + fn
        n = fp + tn
        nan_tensor = torch.tensor(float("nan"), device=confusion_matrix.device)
        if self.metric == "tpr":
            numerator, denominator = tp, p
        elif self.metric == "tnr":
            numerator, denominator = tn, n
        elif self.metric == "ppv":
            numerator, denominator = tp, (tp + fp)
        elif self.metric == "npv":
            numerator, denominator = tn, (tn + fn)
        elif self.metric == "fnr":
            numerator, denominator = fn, p
        elif self.metric == "fpr":
            numerator, denominator = fp, n
        elif self.metric == "fdr":
            numerator, denominator = fp, (fp + tp)
        elif self.metric == "for":
            numerator, denominator = fn, (fn + tn)
        elif self.metric == "pt":
            tpr = torch.where(p > 0, tp / p, nan_tensor)
            tnr = torch.where(n > 0, tn / n, nan_tensor)
            numerator = torch.sqrt(tpr * (1.0 - tnr)) + tnr - 1.0
            denominator = tpr + tnr - 1.0
        elif self.metric == "ts":
            numerator, denominator = tp, (tp + fn + fp)
        elif self.metric == "acc":
            numerator, denominator = (tp + tn), (p + n)
        elif self.metric == "ba":
            tpr = torch.where(p > 0, tp / p, nan_tensor)
            tnr = torch.where(n > 0, tn / n, nan_tensor)
            numerator, denominator = (tpr + tnr), 2.0
        elif self.metric == "f1":
            numerator, denominator = tp * 2.0, (tp * 2.0 + fn + fp)
        elif self.metric == "mcc":
            numerator = tp * tn - fp * fn
            denominator = torch.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))
        elif self.metric == "fm":
            tpr = torch.where(p > 0, tp / p, nan_tensor)
            ppv = torch.where((tp + fp) > 0, tp / (tp + fp), nan_tensor)
            numerator = torch.sqrt(ppv * tpr)
            denominator = 1.0
        elif self.metric == "bm":
            tpr = torch.where(p > 0, tp / p, nan_tensor)
            tnr = torch.where(n > 0, tn / n, nan_tensor)
            numerator = tpr + tnr - 1.0
            denominator = 1.0
        elif self.metric == "mk":
            ppv = torch.where((tp + fp) > 0, tp / (tp + fp), nan_tensor)
            npv = torch.where((tn + fn) > 0, tn / (tn + fn), nan_tensor)
            numerator = ppv + npv - 1.0
            denominator = 1.0
        else:
            raise NotImplementedError("the metric is not implemented.")
        index = numerator / (denominator + self.eps)

        if self.reduction == 'mean':
            index = index.mean()
        elif self.reduction == 'sum':
            index = index.mean()
        elif self.reduction == 'none':
            return index
        return index

    def _check_conf_matrix_format(self, confusion_matrix):
        input_dim = confusion_matrix.ndimension()
        if input_dim == 1:
            confusion_matrix = confusion_matrix.unsqueeze(dim=0)
        if confusion_matrix.shape[-1] != 4:
            raise ValueError("the size of the last dimension of confusion_matrix should be 4.")
        return confusion_matrix

    def _check_confusion_matrix_metric_name(self, metric_name):
        """
        This function is used to check and simplify compliant names.
        There are many metrics related to confusion matrices, some of which have more than
        one name. This function is used to check and simplify names.
        """
        metric_name = metric_name.replace(" ", "_")
        metric_name = metric_name.lower()
        if metric_name in ["sensitivity", "recall", "hit_rate", "true_positive_rate", "tpr"]:
            return "tpr"
        if metric_name in ["specificity", "selectivity", "true_negative_rate", "tnr"]:
            return "tnr"
        if metric_name in ["precision", "positive_predictive_value", "ppv"]:
            return "ppv"
        if metric_name in ["negative_predictive_value", "npv"]:
            return "npv"
        if metric_name in ["miss_rate", "false_negative_rate", "fnr"]:
            return "fnr"
        if metric_name in ["fall_out", "false_positive_rate", "fpr"]:
            return "fpr"
        if metric_name in ["false_discovery_rate", "fdr"]:
            return "fdr"
        if metric_name in ["false_omission_rate", "for"]:
            return "for"
        if metric_name in ["prevalence_threshold", "pt"]:
            return "pt"
        if metric_name in ["threat_score", "critical_success_index", "ts", "csi"]:
            return "ts"
        if metric_name in ["accuracy", "acc"]:
            return "acc"
        if metric_name in ["balanced_accuracy", "ba"]:
            return "ba"
        if metric_name in ["f1_score", "f1"]:
            return "f1"
        if metric_name in ["matthews_correlation_coefficient", "mcc"]:
            return "mcc"
        if metric_name in ["fowlkes_mallows_index", "fm"]:
            return "fm"
        if metric_name in ["informedness", "bookmaker_informedness", "bm", "youden_index", "youden"]:
            return "bm"
        if metric_name in ["markedness", "deltap", "mk"]:
            return "mk"
        raise NotImplementedError("the metric is not implemented.")

if __name__=="__main__":
    metric_name = "acc"

    confusion_metric = ConfusionMatrixs3D(metric_name)
    batch_size = 2
    num_channels = 3
    depth = 2
    height = 2
    width = 2
    y_pred = torch.randint(0, num_channels, (batch_size, num_channels, depth, height, width))
    y_true = torch.randint(0, num_channels, (batch_size, num_channels, depth, height, width))
    result = confusion_metric(y_pred, y_true)

    print(f"{metric_name}: {result}")


from .io import (
    ImagesReader, imwrite, imshowplt, StackedCV2, StackedImages, Stackedplt,
    Stackedtorch, plot_line, bar_chart, scatter_plot, url2cvImage
)
from .color import *
from .mediapipe import Hands
from .video_utils import FPS, VideoCap, Timer, Mp4toGif
from .printf import set_logger, printlog, colorful, colorstr
from .utils import RemoveInboxes, cvt8png, cvtMask, imageContentMove




"""
ShuffleNet: An Extremely Efficient Convolutional Neural Network for Mobile Devices
    <https://arxiv.org/pdf/1707.01083.pdf>
ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design
    <https://arxiv.org/pdf/1807.11164.pdf>
"""
import torch
import torch.nn as nn

__all__ = ["ShuffleNetV1", "shufflenet_v1_g1", "shufflenet_v1_g2", "shufflenet_v1_g3",
           "shufflenet_v1_g4", "shufflenet_v1_g8", "ShuffleNetV2", "shufflenet_v2_x0_5",
           "shufflenet_v2_x1_0", "shufflenet_v2_x1_5", "shufflenet_v2_x2_0",]

cfg_with_shufflenet = {
    'g1': [144, 288, 576],
    'g2': [200, 400, 800],
    'g3': [240, 480, 960],
    'g4': [272, 544, 1088],
    'g8': [384, 768, 1536],
    'x0_5': [24, 48, 96, 192, 1024],
    'x1_0': [24, 116, 232, 464, 1024],
    'x1_5': [24, 176, 352, 704, 1024],
    'x2_0': [24, 244, 488, 976, 2048],
}

def channel_shuffle(x, groups):
    """Shufflenet uses channel shuffling"""
    batchsize, num_channels, height, width = x.size()
    channels_per_group = num_channels // groups
    x = x.view(batchsize, groups, channels_per_group, height, width)   # reshape
    x = torch.transpose(x, 1, 2).contiguous()
    x = x.view(batchsize, num_channels, height, width)  # flatten
    return x

class shufflenet_units(nn.Module):
    """可参考 <https://arxiv.org/pdf/1707.01083.pdf> Figure2 (b) and (c)"""
    def __init__(self, in_channels, out_channels, stride, groups):
        super(shufflenet_units, self).__init__()
        mid_channels = out_channels // 4
        self.stride = stride
        self.groups = 1 if in_channels == 24 else groups

        self.GConv1 = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=1, stride=1, groups=self.groups, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True)
        )

        self.DWConv = nn.Sequential(
            nn.Conv2d(mid_channels, mid_channels, kernel_size=3, stride=self.stride, padding=1, groups=self.groups, bias=False),
            nn.BatchNorm2d(mid_channels)
        )

        self.GConv2 = nn.Sequential(
            nn.Conv2d(mid_channels, out_channels, kernel_size=1, stride=1, groups=self.groups, bias=False),
            nn.BatchNorm2d(out_channels)
        )
        self.shortcut = nn.AvgPool2d(kernel_size=3, stride=2, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.GConv1(x)
        out = channel_shuffle(out, groups=self.groups)
        out = self.DWConv(out)
        out = self.GConv2(out)
        if self.stride == 2:
            short_out = self.shortcut(x)
            out = torch.cat([out, short_out], dim=1)
        return self.relu(out)


class ShuffleNetV1(nn.Module):
    """
    参考 <https://arxiv.org/pdf/1707.01083.pdf> Table 1 实现
    根据论文所述 —— 组数越小表示性能越好
    """
    def __init__(self, groups, stages_out_channels, num_classes=1000, repeat_layers=(4, 8, 4)):
        super(ShuffleNetV1, self).__init__()
        self.groups = groups
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 24, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(24),
            nn.ReLU(inplace=True),
        )
        self.max_pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.stage2 = self._make_layers(24, stages_out_channels[0], repeat_layers[0], groups)
        self.stage3 = self._make_layers(stages_out_channels[0], stages_out_channels[1], repeat_layers[1], groups)
        self.stage4 = self._make_layers(stages_out_channels[1], stages_out_channels[2], repeat_layers[2], groups)

        self.globalpool = nn.AvgPool2d(kernel_size=7, stride=1)
        self.fc = nn.Linear(stages_out_channels[2], num_classes)

    def _make_layers(self, in_channels, out_channels, repeat_number, groups):
        layers = []
        # 不同的 stage 阶段在第一步都是 stride = 2
        layers.append(shufflenet_units(in_channels, out_channels - in_channels, 2, groups))
        in_channels = out_channels
        for i in range(repeat_number - 1):
            layers.append(shufflenet_units(in_channels, out_channels, 1, groups))
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.max_pool(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)
        x = self.globalpool(x)
        x = torch.flatten(x, 1)
        out = self.fc(x)
        return out

################################### 以下为 ShuffleNetV2 ###############################################

class Improved_shufflenet_units(nn.Module):
    def __init__(self, inp, oup, stride):
        super().__init__()
        if not (1 <= stride <= 3):
            raise ValueError("illegal stride value")
        self.stride = stride
        branch_features = oup // 2
        if (self.stride == 1) and (inp != branch_features << 1):
            raise ValueError(
                f"Invalid combination of stride {stride}, inp {inp} and oup {oup} values. If stride == 1 then inp should be equal to oup // 2 << 1."
            )

        self.branch1 = nn.Sequential(
            # DWConv
            nn.Conv2d(inp, inp, kernel_size=3, stride=self.stride, padding=1, bias=False, groups=inp),
            nn.BatchNorm2d(inp),
            nn.Conv2d(inp, branch_features, kernel_size=1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(branch_features),
            nn.ReLU(inplace=True),
        )
        self.branch2 = nn.Sequential(
            nn.Conv2d(
                inp if (self.stride > 1) else branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.BatchNorm2d(branch_features),
            nn.ReLU(inplace=True),
            # DWConv
            nn.Conv2d(branch_features, branch_features, kernel_size=3, stride=self.stride, padding=1, bias=False, groups=branch_features),
            nn.BatchNorm2d(branch_features),
            nn.Conv2d(branch_features, branch_features, kernel_size=1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(branch_features),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        if self.stride == 1:
            x1, x2 = x.chunk(2, dim=1)
            out = torch.cat((x1, self.branch2(x2)), dim=1)
        else:
            out = torch.cat((self.branch1(x), self.branch2(x)), dim=1)

        out = channel_shuffle(out, 2)

        return out


class ShuffleNetV2(nn.Module):
    def __init__(self,
                 stages_out_channels,
                 num_classes=1000,
                 improved_units=Improved_shufflenet_units,
                 repeat_layers=(4, 8, 4)):
        super().__init__()
        if len(repeat_layers) != 3:
            raise ValueError("expected stages_repeats as list of 3 positive ints")
        if len(stages_out_channels) != 5:
            raise ValueError("expected stages_out_channels as list of 5 positive ints")
        self._stage_out_channels = stages_out_channels

        input_channels = 3
        output_channels = self._stage_out_channels[0]
        self.conv1 = nn.Sequential(
            nn.Conv2d(input_channels, output_channels, 3, 2, 1, bias=False),
            nn.BatchNorm2d(output_channels),
            nn.ReLU(inplace=True),
        )
        input_channels = output_channels

        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        # Static annotations for mypy
        self.stage2: nn.Sequential
        self.stage3: nn.Sequential
        self.stage4: nn.Sequential
        stage_names = [f"stage{i}" for i in [2, 3, 4]]
        for name, repeats, output_channels in zip(stage_names, repeat_layers, self._stage_out_channels[1:]):
            seq = [improved_units(input_channels, output_channels, 2)]
            for i in range(repeats - 1):
                seq.append(improved_units(output_channels, output_channels, 1))
            setattr(self, name, nn.Sequential(*seq))
            input_channels = output_channels

        output_channels = self._stage_out_channels[-1]
        self.conv5 = nn.Sequential(
            nn.Conv2d(input_channels, output_channels, 1, 1, 0, bias=False),
            nn.BatchNorm2d(output_channels),
            nn.ReLU(inplace=True),
        )
        self.fc = nn.Linear(output_channels, num_classes)

    def forward(self, x):
        x = self.conv1(x)
        x = self.maxpool(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)
        x = self.conv5(x)
        x = x.mean([2, 3])  # globalpool
        x = self.fc(x)
        return x

#----------------------------------------------------------------------------------------------------#
def shufflenet_v1_g1(num_classes):
    return ShuffleNetV1(groups=1,
                        stages_out_channels=cfg_with_shufflenet['g1'],
                        num_classes=num_classes)

def shufflenet_v1_g2(num_classes):
    return ShuffleNetV1(groups=2,
                        stages_out_channels=cfg_with_shufflenet['g2'],
                        num_classes=num_classes)

def shufflenet_v1_g3(num_classes):
    return ShuffleNetV1(groups=3,
                        stages_out_channels=cfg_with_shufflenet['g3'],
                        num_classes=num_classes)

def shufflenet_v1_g4(num_classes):
    return ShuffleNetV1(groups=4,
                        stages_out_channels=cfg_with_shufflenet['g4'],
                        num_classes=num_classes)

def shufflenet_v1_g8(num_classes):
    return ShuffleNetV1(groups=8,
                        stages_out_channels=cfg_with_shufflenet['g8'],
                        num_classes=num_classes)

def shufflenet_v2_x0_5(num_classes):
    return ShuffleNetV2(cfg_with_shufflenet['x0_5'],
                        num_classes=num_classes)

def shufflenet_v2_x1_0(num_classes):
    return ShuffleNetV2(cfg_with_shufflenet['x1_0'],
                        num_classes=num_classes)

def shufflenet_v2_x1_5(num_classes):
    return ShuffleNetV2(cfg_with_shufflenet['x1_5'],
                        num_classes=num_classes)

def shufflenet_v2_x2_0(num_classes):
    return ShuffleNetV2(cfg_with_shufflenet['x2_0'],
                        num_classes=num_classes)


if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = shufflenet_v2_x2_0(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # shufflenet_v1_g1 Total params: 1,880,296
    # shufflenet_v1_g2 Total params: 1,834,852
    # shufflenet_v1_g3 Total params: 1,781,032
    # shufflenet_v1_g4 Total params: 1,734,136
    # shufflenet_v1_g8 Total params: 1,797,304
    # shufflenet_v2_x0_5 Total params: 345,892
    # shufflenet_v2_x1_0 Total params: 1,257,704
    # shufflenet_v2_x1_5 Total params: 2,482,724
    # shufflenet_v2_x2_0 Total params: 5,353,192


import numpy as np
import torch
from torch import nn
from torch.nn import init



class PSA(nn.Module):

    def __init__(self, channel=512,reduction=4,S=4):
        super().__init__()
        self.S=S

        self.convs=[]
        for i in range(S):
            self.convs.append(nn.Conv2d(channel//S,channel//S,kernel_size=2*(i+1)+1,padding=i+1))

        self.se_blocks=[]
        for i in range(S):
            self.se_blocks.append(nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                nn.Conv2d(channel//S, channel // (S*reduction),kernel_size=1, bias=False),
                nn.ReLU(inplace=True),
                nn.Conv2d(channel // (S*reduction), channel//S,kernel_size=1, bias=False),
                nn.Sigmoid()
            ))

        self.softmax=nn.Softmax(dim=1)


    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, x):
        b, c, h, w = x.size()

        #Step1:SPC module
        SPC_out=x.view(b,self.S,c//self.S,h,w) #bs,s,ci,h,w
        for idx,conv in enumerate(self.convs):
            SPC_out[:,idx,:,:,:]=conv(SPC_out[:,idx,:,:,:])

        #Step2:SE weight
        se_out=[]
        for idx,se in enumerate(self.se_blocks):
            se_out.append(se(SPC_out[:,idx,:,:,:]))
        SE_out=torch.stack(se_out,dim=1)
        SE_out=SE_out.expand_as(SPC_out)

        #Step3:Softmax
        softmax_out=self.softmax(SE_out)

        #Step4:SPA
        PSA_out=SPC_out*softmax_out
        PSA_out=PSA_out.view(b,-1,h,w)

        return PSA_out


if __name__ == '__main__':
    input=torch.randn(50,512,7,7)
    psa = PSA(channel=512,reduction=8)
    output=psa(input)
    a=output.view(-1).sum()
    a.backward()
    print(output.shape)




import torch
import torch.nn as nn


class SimAM(torch.nn.Module):
    def __init__(self, channels = None, e_lambda = 1e-4):
        super(SimAM, self).__init__()

        self.activaton = nn.Sigmoid()
        self.e_lambda = e_lambda

    def __repr__(self):
        s = self.__class__.__name__ + '('
        s += ('lambda=%f)' % self.e_lambda)
        return s

    @staticmethod
    def get_module_name():
        return "simam"

    def forward(self, x):

        b, c, h, w = x.size()

        n = w * h - 1

        x_minus_mu_square = (x - x.mean(dim=[2,3], keepdim=True)).pow(2)
        y = x_minus_mu_square / (4 * (x_minus_mu_square.sum(dim=[2,3], keepdim=True) / n + self.e_lambda)) + 0.5

        return x * self.activaton(y)

if __name__ == '__main__':
    input=torch.randn(3, 64, 7, 7)
    model = SimAM(64)
    outputs = model(input)
    print(outputs.shape)


import torch
import torch.nn as nn
import torch.nn.functional as F

class BasicUNet3D(nn.Module):
    """
    A simple 3D Unet, adapted from a 2D Unet from https://github.com/milesial/Pytorch-UNet/tree/master/unet
    Args:
        in_channels: number of input channels; 3 for RGB, 1 for grayscale input
        num_classes: number of output channels/classes
        width_multiplier: how much 'wider' your UNet should be compared with a standard UNet
                default is 1;, meaning 32 -> 64 -> 128 -> 256 -> 512 -> 256 -> 128 -> 64 -> 32
                higher values increase the number of kernels pay layer, by that factor
        trilinear: use trilinear interpolation to upsample; if false, 3D convtranspose
                layers will be used instead
        use_ds_conv: if True, we use depthwise-separable convolutional layers. in my
                experience, this is of little help. This appears to be because with 3D
                data, the vast vast majority of GPU RAM is the input data/labels, not
                the params, so little VRAM is saved by using ds_conv, and yet performance suffers."""
    def __init__(self, in_channels, num_classes, width_multiplier=1, trilinear=True, use_ds_conv=False):
        super(BasicUNet3D, self).__init__()
        _channels = (32, 64, 128, 256, 512)
        self.in_channels = in_channels
        self.num_classes = num_classes
        self.channels = [int(c*width_multiplier) for c in _channels]
        self.trilinear = trilinear
        self.convtype = DepthwiseSeparableConv3d if use_ds_conv else nn.Conv3d

        self.inc = DoubleConv(in_channels, self.channels[0], conv_type=self.convtype)
        self.down1 = Down(self.channels[0], self.channels[1], conv_type=self.convtype)
        self.down2 = Down(self.channels[1], self.channels[2], conv_type=self.convtype)
        self.down3 = Down(self.channels[2], self.channels[3], conv_type=self.convtype)
        factor = 2 if trilinear else 1
        self.down4 = Down(self.channels[3], self.channels[4] // factor, conv_type=self.convtype)
        self.up1 = Up(self.channels[4], self.channels[3] // factor, trilinear)
        self.up2 = Up(self.channels[3], self.channels[2] // factor, trilinear)
        self.up3 = Up(self.channels[2], self.channels[1] // factor, trilinear)
        self.up4 = Up(self.channels[1], self.channels[0], trilinear)
        self.outc = OutConv(self.channels[0], num_classes)

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return logits


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""
    def __init__(self, in_channels, out_channels, conv_type=nn.Conv3d, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            conv_type(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(mid_channels),
            nn.ReLU(inplace=True),
            conv_type(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels, conv_type=nn.Conv3d):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool3d(2),
            DoubleConv(in_channels, out_channels, conv_type=conv_type)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, trilinear=True):
        super().__init__()

        # if trilinear, use the normal convolutions to reduce the number of channels
        if trilinear:
            self.up = nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, mid_channels=in_channels // 2)
        else:
            self.up = nn.ConvTranspose3d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)


    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv3d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)


class DepthwiseSeparableConv3d(nn.Module):
    def __init__(self, nin, nout, kernel_size, padding, kernels_per_layer=1):
        super(DepthwiseSeparableConv3d, self).__init__()
        self.depthwise = nn.Conv3d(nin, nin * kernels_per_layer, kernel_size=kernel_size, padding=padding, groups=nin)
        self.pointwise = nn.Conv3d(nin * kernels_per_layer, nout, kernel_size=1)

    def forward(self, x):
        out = self.depthwise(x)
        out = self.pointwise(out)
        return out

if __name__ == '__main__':
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    # B x C x Z x Y x X
    # 4 x 1 x 64 x 64 x 64
    input = torch.ones(1, 1, 32, 32, 32).to(device)
    net = BasicUNet3D(in_channels=1, num_classes=3)
    net = net.to(device)
    out = net(input)
    print(out.shape)
    torchsummary.summary(net, input_size=(1, 32, 32, 32))


"""
Copyright (c) 2023, Auorui.
All rights reserved.

"""
from .point_line import *
from .arithmetic import *
from .constant import *
from .function import *
from .Numpy import rand, normal, gaussian2d



"""
A2-Nets: Double Attention Networks
Original paper addresshttps: <https://arxiv.org/pdf/1810.11579.pdf>
"""
import torch
from torch import nn
from torch.nn import init
from torch.nn import functional as F

class DoubleAttention(nn.Module):
    def __init__(self, in_channels, c_m, c_n, reconstruct=True):
        super().__init__()
        self.in_channels = in_channels
        self.reconstruct = reconstruct
        self.c_m = c_m
        self.c_n = c_n
        self.convA = nn.Conv2d(in_channels, c_m, 1)
        self.convB = nn.Conv2d(in_channels, c_n, 1)
        self.convV = nn.Conv2d(in_channels, c_n, 1)
        if self.reconstruct:
            self.conv_reconstruct = nn.Conv2d(c_m, in_channels, kernel_size=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, x):
        b, c, h, w = x.shape
        assert c == self.in_channels
        A = self.convA(x)  # b, c_m, h, w
        B = self.convB(x)  # b, c_n, h, w
        V = self.convV(x)  # b, c_n, h, w
        tmp_a = A.view(b, self.c_m, -1)
        attention_maps = F.softmax(B.view(b, self.c_n, -1), dim=-1)
        attention_vectors = F.softmax(V.view(b, self.c_n, -1), dim=-1)
        # 第一步：特征提取
        global_descriptors = torch.bmm(tmp_a, attention_maps.permute(0, 2, 1))  # b.c_m, c_n
        # 第二步：特征分布
        tmp_z = global_descriptors.matmul(attention_vectors)  # b, c_m, h*w
        tmp_z = tmp_z.view(b, self.c_m, h, w)  # b,c_m,h,w
        if self.reconstruct:
            tmp_z = self.conv_reconstruct(tmp_z)

        return tmp_z


if __name__ == '__main__':
    block = DoubleAttention(64, 128, 128)
    input = torch.rand(1, 64, 64, 64)
    output = block(input)
    print(output.shape)


"""
This module provides 3D neural networks for DSCNet, UNet, and VNet

Copyright (c) 2024, Auorui.
All rights reserved.
time 2024-04-28 The last weekend of April.
"""
from .basic_unet import BasicUNet3D
from .vent import VNet3D
from .dscnet import DSCNet3D


from .semantic import (
    Miou,
    Recall,
    Precision,
    F1Score,
    DiceCoefficient,
    Accuracy,
    SegmentationIndex,
    AIU
)

from .classification import (
    accuracy_all_classes,
    cls_matrix,
    BinaryConfusionMatrix,
    MulticlassConfusionMatrix,
    ConfusionMatrixs,
    ModelIndex,
    calculate_metrics,
    MultiLabelConfusionMatrix
)

from .segment3d import (
    PPV,
    DiceMetric3d,
    MeanIoU3d,
    JaccardMetric3d,
    HausdorffDistanceMetric3d
)

from .indexutils import (
    hd, hd95,
    _surface_distances,
    ignore_background,
)

from .medical_index import (
    ConfusionMatrixs3D,
    get_confusion_matrix_3d,
    get_confusion_matrix_3d_np,
    get_confusion_matrix_3d_torch
)


"""
Code originates from: https://github.com/mindspore-courses/External-Attention-MindSpore/blob/main/model/attention/UFOAttention.py
Refer to "UFO-ViT: High Performance Linear Vision Transformer without Softmax"
Original paper addresshttps: https://arxiv.org/pdf/2109.14382.pdf
"""
import torch
from torch import nn
from torch.nn import init

class UFOAttention(nn.Module):
    def __init__(self, d_model, d_k=None, d_v=None, h=1, dropout=0.1):
        super(UFOAttention, self).__init__()
        if d_k is None:
            d_k = d_model
        if d_v is None:
            d_v = d_model

        self.fc_q = nn.Linear(d_model, h * d_k)
        self.fc_k = nn.Linear(d_model, h * d_k)
        self.fc_v = nn.Linear(d_model, h * d_v)
        self.fc_o = nn.Linear(h * d_v, d_model)
        self.dropout = nn.Dropout(p=dropout)
        self.gamma = nn.Parameter(torch.randn((1, h, 1, 1)))

        self.d_model = d_model
        self.d_k = d_k
        self.d_v = d_v
        self.h = h
        for m in self.modules():
            if isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, queries, keys=None, values=None):
        if keys is None:
            keys = queries
        if values is None:
            values = keys

        B, N = queries.shape[:2]
        q = self.fc_q(queries).view(B, N, self.h, self.d_k).permute(0, 2, 1, 3)
        k = self.fc_k(keys).view(B, N, self.h, self.d_k).permute(0, 2, 3, 1)
        v = self.fc_v(values).view(B, N, self.h, self.d_v).permute(0, 2, 1, 3)

        kv = torch.matmul(k, v)
        kv_norm = self.XNorm(kv, self.gamma)
        q_norm = self.XNorm(q, self.gamma)
        out = torch.matmul(q_norm, kv_norm).permute(0, 2, 1, 3).reshape(B, N, self.h * self.d_v)
        return self.fc_o(out)

    def XNorm(self, x, gamma):
        norm_tensor = torch.norm(x, 2, -1, keepdim=True)
        return x * gamma / norm_tensor

if __name__ == '__main__':
    dummy_input = torch.randn((25, 49, 512))
    ufo = UFOAttention(d_model=512, h=8)
    output = ufo(dummy_input)
    print(output.shape)




import os
import torch
import random
import numpy as np
from pathlib import Path
from PIL import Image
from pyzjr.core import to_2tuple
from torch.utils.data import Dataset
from pyzjr.core.__tensor import img2tensor, label2tensor

class BaseDataset(Dataset):
    """A simple dataset class with the same functionality as torch.utils.data.Dataset"""
    def __init__(self):
        pass

    def __getitem__(self, idx):
        pass

    def __len__(self):
        pass

    def list_files(self, imagefolder, image_extensions=('.jpg', '.jpeg', '.png', '.bmp')):
        """
        返回给定目录中的所有图像文件
        """
        return [
            f for f in Path(imagefolder).iterdir()
            if f.is_file() and not f.name.startswith(".") and f.suffix.lower() in image_extensions
        ]

    def list_dirs(self, imagefolder):
        """
        返回给定目录中的所有目录
        """
        return [f for f in Path(imagefolder).iterdir() if f.is_dir()]

    def rand(self, a=0., b=1.):
        """
        生成范围在 [a, b) 之间的随机数
        """
        return np.random.rand() * (b - a) + a

    def osmkdirs(self, *args):
        """
        如果不存在的话,创建多个目录
        """
        for path in args:
            if not os.path.exists(path):
                os.makedirs(path)

    def preprocess_input(self, image):
        """
        将图像转换为浮点数并进行归一化
        """
        image = np.array(image).astype(np.float32)
        image /= 255.0
        return image

    def check_input(self, input, classes=(float, int, int, str, list, tuple, np.ndarray, torch.Tensor)):
        """
        检查数据类型是否符合指定的类别
        classes: float, int, int, str, list, tuple, np.ndarray, torch.Tensor
        """
        return isinstance(input, classes)

    def Shuffle(self, files):
        """
        随机打乱文件列表
        """
        assert self.check_input(files, classes=(list, tuple)), "Input must be a list or tuple"
        random.shuffle(files)
        return files

    def load_txt_path_or_list(self, txt_path):
        """
        加载txt文件或列表,按行读取,去掉'/n'
        """
        path_list = []
        if self.check_input(txt_path, str):  # Check if txt_path is a string (path to a file)
            with open(txt_path, 'r') as file:
                lines = file.readlines()
                path_list.extend(line.strip() for line in lines)
        elif self.check_input(txt_path, list):  # Check if txt_path is a list
            path_list.extend(item.strip() for item in txt_path)
        else:
            raise ValueError(f"Invalid input type. {txt_path} should be a string (file path) or a list.")
        return path_list

    def Paste2center(self, image, label, input_shape):
        """
        将图像和标签粘贴到指定输入形状的中心位置。

        :param image: 输入图像 (PIL.Image)
        :param label: 输入标签图像 (PIL.Image)
        :param input_shape: 目标输入形状 (tuple)，形如 (width, height)
        :return: 调整后的图像和标签
        """
        input_shape = to_2tuple(input_shape)
        h, w = input_shape
        iw, ih = image.size
        scale = min(w / iw, h / ih)
        nw = int(iw * scale)
        nh = int(ih * scale)

        image = image.resize((nw, nh), Image.BICUBIC)
        new_image = Image.new('RGB', (w, h), (128, 128, 128))
        new_image.paste(image, ((w - nw) // 2, (h - nh) // 2))

        label = label.resize((nw, nh), Image.NEAREST)
        new_label = Image.new('L', (w, h), (0))
        new_label.paste(label, ((w - nw) // 2, (h - nh) // 2))
        return new_image, new_label

    def imlabel2tensor(self, im, label, num_classes, sigmoid, totensor=False):
        im = img2tensor(im, totensor)
        label = label2tensor(label, num_classes, sigmoid, totensor)
        return im, label


"""
收录或复现的其他改进的卷积
"""
from .Snake_Conv import DSConv
from .Depthwise_Conv import DepthSepConv, DWConv, PWConv
from .Partial_Conv import PartialConv
from .Ref_Conv import RefConv

__all__ = ["DSConv", "DepthSepConv", "PartialConv", "RefConv", "DWConv", "PWConv"]


import os
import glob
import nibabel as nib
import numpy as np

def get_nii_path(folder_path):
    """
    在给定的文件夹路径中搜索并返回所有.nii.gz文件的路径。
    Args:
        folder_path (str): 要搜索的文件夹路径。

    Return:
        list: 包含所有找到的.nii.gz文件路径的列表。
    """
    image_files = []
    for extension in ['*.nii.gz']:
        image_files.extend(glob.glob(os.path.join(folder_path, extension)))
    return image_files

def nii_load(path):
    """
    读取NIfTI格式的图像文件。 ---> (w, h, d) [ x, y, z ]
    Args:
        path: str, NIfTI图像文件的路径

    Return:
        data: numpy.ndarray, 包含图像数据的三维数组。
        affine: numpy.ndarray, 图像的仿射矩阵，用于将图像坐标转换为世界坐标。
        header: nibabel.Nifti1Header, NIfTI图像的头文件信息。
    """
    nii_image = nib.load(path)
    data = nii_image.get_fdata()
    affine = nii_image.affine
    header = nii_image.header
    return data, affine, header

def nii_write(outputs_np, affine, header, filename_out):
    """
    将给定的 Numpy 数组保存为 NIfTI 格式的图像文件。

    Args:
        outputs_np: numpy.ndarray, 要保存的图像数据。
        affine: numpy.ndarray, 图像的仿射矩阵，描述了图像的空间位置和定向信息。
        header: numpy.ndarray, 图像的头部信息
        filename_out: str, 输出文件的路径，包括文件名及其扩展名（如'.nii.gz'）。
    """
    outputs_nib = nib.Nifti1Image(outputs_np, affine, header)
    outputs_nib.header['qform_code'] = 1
    outputs_nib.header['sform_code'] = 0
    outputs_nib.to_filename(filename_out)

def center_crop_nii(data, crop_shape):
    """
    对 NIfTI 图像进行中心裁剪。
    Args:
        data: numpy.ndarray, NIfTI 图像数据。
        crop_shape: tuple, 裁剪后的图像形状 (depth, height, width)。

    Returns:
        cropped_data: numpy.ndarray, 裁剪后的图像数据。
    """
    start_x = (data.shape[2] - crop_shape[2]) // 2
    start_y = (data.shape[1] - crop_shape[1]) // 2
    start_z = (data.shape[0] - crop_shape[0]) // 2
    cropped_data = data[start_z:start_z + crop_shape[0],
                   start_y:start_y + crop_shape[1],
                   start_x:start_x + crop_shape[2]]
    return cropped_data.astype(np.uint8)


def center_scale_nii(data, scale_factor):
    """
    对给定的NIfTI格式数据进行中心缩放。

    参数：
        data (numpy.ndarray)：NIfTI数据的Numpy数组表示。
        scale_factor (float)：缩放因子，用于指定缩放的比例。

    返回：
        numpy.ndarray：经过中心缩放后的数据。

    注意：
        这个函数假设输入的数据是3D的，并且是一个Numpy数组。你应该调用 nii_load 读取
        缩放因子应为正数，表示缩小（<1）或放大（>1）的比例。
    """
    new_shape = np.round(np.array(data.shape) * scale_factor).astype(int)
    offset = (np.array(data.shape) - new_shape) // 2
    scaled_data = data[offset[0]:offset[0]+new_shape[0],
                  offset[1]:offset[1]+new_shape[1],
                  offset[2]:offset[2]+new_shape[2]]
    return scaled_data.astype(np.uint8)


import torch
import numpy as np
from scipy.ndimage import _ni_support, binary_erosion, distance_transform_edt, generate_binary_structure

def _surface_distances(y_pred, y, voxelspacing=None, connectivity=1):
    """
    计算两个二值对象中表面体素之间的距离
    """
    y_pred = np.atleast_1d(y_pred.astype(np.bool_))
    y = np.atleast_1d(y.astype(np.bool_))
    if voxelspacing is not None:
        voxelspacing = _ni_support._normalize_sequence(voxelspacing, y_pred.ndim)
        voxelspacing = np.asarray(voxelspacing, dtype=np.float64)
        if not voxelspacing.flags.contiguous:
            voxelspacing = voxelspacing.copy()

    # binary structure
    footprint = generate_binary_structure(y_pred.ndim, connectivity)

    # test for emptiness
    if 0 == np.count_nonzero(y_pred):
        raise RuntimeError(
            "The first supplied array does not contain any binary object."
        )
    if 0 == np.count_nonzero(y):
        raise RuntimeError(
            "The second supplied array does not contain any binary object."
        )
    result_border = y_pred ^ binary_erosion(y_pred, structure=footprint, iterations=1)
    reference_border = y ^ binary_erosion(
        y, structure=footprint, iterations=1
    )
    dt = distance_transform_edt(~reference_border, sampling=voxelspacing)
    sds = dt[result_border]

    return sds

def hd(y_pred, y, voxelspacing=None, connectivity=1):
    """
    Hausdorff Distance.
    Hausdorff距离衡量了两个点集之间的最大不匹配程度。在此函数中，它用于计算两个二值图像（result和reference）的表面体素之间的最大不匹配距离。

    Args:
        y_pred (array_like): 预测的二值图像，形状与参考图像（y）相同。
        y (array_like): 参考的二值图像，形状与预测图像（y_pred）相同。
        voxelspacing (array_like, optional): 形状为（N，）的数组，表示每个维度上的体素间距。默认为None。
        connectivity (int, optional): 表面距离计算中使用的连接性。默认为1，表示3D中的26连通（在3D情况下）或2D中的8连通（在2D情况下）。

    Returns:
        float: Hausdorff距离，即两个图像表面体素之间的最大不匹配距离。
    """
    hd1 = _surface_distances(y_pred, y, voxelspacing, connectivity).max()
    hd2 = _surface_distances(y, y_pred, voxelspacing, connectivity).max()
    hausdorff = max(hd1, hd2)
    return hausdorff

def hd95(y_pred, y, voxelspacing=None, connectivity=1):
    """
    计算 Hausdorff 距离的 95% 的数据点的值
    """
    hausdorff = hd(y_pred, y, voxelspacing, connectivity)
    return np.percentile(hausdorff, 95)

def ignore_background(y_pred, y):
    """
    Used to remove the background (first channel) of "yyred" and "y".
    Args:
        y_pred: predictions. As for classification tasks, `y_pred` should has the shape [BN] where N is larger than 1. As for segmentation tasks,
        the shape should be [BNHW] or [BNHWD].
        y: ground truth, the first dim is batch.
    """
    y = y[:, 1:] if y.shape[1] > 1 else y
    y_pred = y_pred[:, 1:] if y_pred.shape[1] > 1 else y_pred
    return y_pred, y


if __name__=="__main__":
    y_pred = np.random.randint(0, 3, size=(100, 100))
    y = np.random.randint(0, 3, size=(100, 100))
    hausdorff_distance = hd95(y_pred, y)
    print("Hausdorff Distance:", hausdorff_distance)

    batch_size = 1
    num_channels = 3
    depth = 2
    height = 2
    width = 2

    y_pred = torch.rand(batch_size, num_channels, depth, height, width)
    y_true = torch.rand(batch_size, num_channels, depth, height, width)
    print("Shape of y_pred:", y_pred.shape)
    print("Shape of y_true:", y_true.shape)


"""
EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks
     <https://arxiv.org/abs/1905.11946>
EfficientNetV2: Smaller Models and Faster Training
    <https://arxiv.org/abs/2104.00298>
"""
import copy
import math
import warnings
from dataclasses import dataclass
from functools import partial
from typing import Callable, Sequence

import torch
import torch.nn as nn
from torchvision.ops import StochasticDepth, Conv2dNormActivation
from pyzjr.Models.attention import SqueezeExcitation
from pyzjr.Models._utils import _make_divisible


__all__ = [
    "EfficientNet",
    "efficientnet_b0",
    "efficientnet_b1",
    "efficientnet_b2",
    "efficientnet_b3",
    "efficientnet_b4",
    "efficientnet_b5",
    "efficientnet_b6",
    "efficientnet_b7",
    "efficientnet_v2_s",
    "efficientnet_v2_m",
    "efficientnet_v2_l",
]

@dataclass
class _MBConvConfig:
    expand_ratio: float
    kernel: int
    stride: int
    input_channels: int
    out_channels: int
    num_layers: int
    block: Callable[..., nn.Module]

    @staticmethod
    def adjust_channels(channels, width_mult, min_value=None):
        return _make_divisible(channels * width_mult, 8, min_value)


class MBConvConfig(_MBConvConfig):
    # Stores information listed at Table 1 of the EfficientNet paper & Table 4 of the EfficientNetV2 paper
    def __init__(
            self,
            expand_ratio,
            kernel,
            stride,
            input_channels,
            out_channels,
            num_layers,
            width_mult=1.0,
            depth_mult=1.0,
            block=None,
    ) -> None:
        input_channels = self.adjust_channels(input_channels, width_mult)
        out_channels = self.adjust_channels(out_channels, width_mult)
        num_layers = self.adjust_depth(num_layers, depth_mult)
        if block is None:
            block = MBConv
        super().__init__(expand_ratio, kernel, stride, input_channels, out_channels, num_layers, block)

    @staticmethod
    def adjust_depth(num_layers: int, depth_mult: float):
        return int(math.ceil(num_layers * depth_mult))


class FusedMBConvConfig(_MBConvConfig):
    # Stores information listed at Table 4 of the EfficientNetV2 paper
    def __init__(
            self,
            expand_ratio,
            kernel,
            stride,
            input_channels,
            out_channels,
            num_layers,
            block=None,
    ):
        if block is None:
            block = FusedMBConv
        super().__init__(expand_ratio, kernel, stride, input_channels, out_channels, num_layers, block)


class MBConv(nn.Module):
    def __init__(
            self,
            cnf: MBConvConfig,
            stochastic_depth_prob,
            norm_layer,
            se_layer=SqueezeExcitation,):
        super().__init__()

        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = cnf.stride == 1 and cnf.input_channels == cnf.out_channels

        layers = []
        activation_layer = nn.SiLU

        # expand
        expanded_channels = cnf.adjust_channels(cnf.input_channels, cnf.expand_ratio)
        if expanded_channels != cnf.input_channels:
            layers.append(
                Conv2dNormActivation(
                    cnf.input_channels,
                    expanded_channels,
                    kernel_size=1,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                )
            )

        # depthwise
        layers.append(
            Conv2dNormActivation(
                expanded_channels,
                expanded_channels,
                kernel_size=cnf.kernel,
                stride=cnf.stride,
                groups=expanded_channels,
                norm_layer=norm_layer,
                activation_layer=activation_layer,
            )
        )

        # squeeze and excitation
        squeeze_channels = max(1, cnf.input_channels // 4)
        layers.append(se_layer(expanded_channels, squeeze_channels, activation=partial(nn.SiLU, inplace=True)))

        # project
        layers.append(
            Conv2dNormActivation(
                expanded_channels, cnf.out_channels, kernel_size=1, norm_layer=norm_layer, activation_layer=None
            )
        )

        self.block = nn.Sequential(*layers)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")
        self.out_channels = cnf.out_channels

    def forward(self, input):
        result = self.block(input)
        if self.use_res_connect:
            result = self.stochastic_depth(result)
            result += input
        return result


class FusedMBConv(nn.Module):
    def __init__(
            self,
            cnf: FusedMBConvConfig,
            stochastic_depth_prob,
            norm_layer,
    ) -> None:
        super().__init__()

        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = cnf.stride == 1 and cnf.input_channels == cnf.out_channels

        layers = []
        activation_layer = nn.SiLU

        expanded_channels = cnf.adjust_channels(cnf.input_channels, cnf.expand_ratio)
        if expanded_channels != cnf.input_channels:
            # fused expand
            layers.append(
                Conv2dNormActivation(
                    cnf.input_channels,
                    expanded_channels,
                    kernel_size=cnf.kernel,
                    stride=cnf.stride,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                )
            )

            # project
            layers.append(
                Conv2dNormActivation(
                    expanded_channels, cnf.out_channels, kernel_size=1, norm_layer=norm_layer, activation_layer=None
                )
            )
        else:
            layers.append(
                Conv2dNormActivation(
                    cnf.input_channels,
                    cnf.out_channels,
                    kernel_size=cnf.kernel,
                    stride=cnf.stride,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                )
            )

        self.block = nn.Sequential(*layers)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")
        self.out_channels = cnf.out_channels

    def forward(self, input):
        result = self.block(input)
        if self.use_res_connect:
            result = self.stochastic_depth(result)
            result += input
        return result


class EfficientNet(nn.Module):
    def __init__(
            self,
            inverted_residual_setting,
            dropout,
            stochastic_depth_prob=0.2,
            num_classes=1000,
            norm_layer=None,
            last_channel=None,
            **kwargs,
    ):
        """
        EfficientNet V1 and V2 main class

        Args:
            inverted_residual_setting: Network structure
            dropout: The droupout probability
            stochastic_depth_prob: The stochastic depth probability
            num_classes: Number of classes
            norm_layer: Module specifying the normalization layer to use
            last_channel: The number of channels on the penultimate layer
        """
        super().__init__()

        if not inverted_residual_setting:
            raise ValueError("The inverted_residual_setting should not be empty")
        elif not (
                isinstance(inverted_residual_setting, Sequence)
                and all([isinstance(s, _MBConvConfig) for s in inverted_residual_setting])
        ):
            raise TypeError("The inverted_residual_setting should be List[MBConvConfig]")

        if "block" in kwargs:
            warnings.warn(
                "The parameter 'block' is deprecated since 0.13 and will be removed 0.15. "
                "Please pass this information on 'MBConvConfig.block' instead."
            )
            if kwargs["block"] is not None:
                for s in inverted_residual_setting:
                    if isinstance(s, MBConvConfig):
                        s.block = kwargs["block"]

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        layers = []

        # building first layer
        firstconv_output_channels = inverted_residual_setting[0].input_channels
        layers.append(
            Conv2dNormActivation(
                3, firstconv_output_channels, kernel_size=3, stride=2, norm_layer=norm_layer, activation_layer=nn.SiLU
            )
        )

        # building inverted residual blocks
        total_stage_blocks = sum(cnf.num_layers for cnf in inverted_residual_setting)
        stage_block_id = 0
        for cnf in inverted_residual_setting:
            stage = []
            for _ in range(cnf.num_layers):
                # copy to avoid modifications. shallow copy is enough
                block_cnf = copy.copy(cnf)

                # overwrite info if not the first conv in the stage
                if stage:
                    block_cnf.input_channels = block_cnf.out_channels
                    block_cnf.stride = 1

                # adjust stochastic depth probability based on the depth of the stage block
                sd_prob = stochastic_depth_prob * float(stage_block_id) / total_stage_blocks

                stage.append(block_cnf.block(block_cnf, sd_prob, norm_layer))
                stage_block_id += 1

            layers.append(nn.Sequential(*stage))

        # building last several layers
        lastconv_input_channels = inverted_residual_setting[-1].out_channels
        lastconv_output_channels = last_channel if last_channel is not None else 4 * lastconv_input_channels
        layers.append(
            Conv2dNormActivation(
                lastconv_input_channels,
                lastconv_output_channels,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=nn.SiLU,
            )
        )

        self.features = nn.Sequential(*layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Sequential(
            nn.Dropout(p=dropout, inplace=True),
            nn.Linear(lastconv_output_channels, num_classes),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                init_range = 1.0 / math.sqrt(m.out_features)
                nn.init.uniform_(m.weight, -init_range, init_range)
                nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.features(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        x = self.classifier(x)

        return x



def _efficientnet_conf(
        arch,
        **kwargs,
):
    if arch.startswith("efficientnet_b"):
        bneck_conf = partial(MBConvConfig, width_mult=kwargs.pop("width_mult"), depth_mult=kwargs.pop("depth_mult"))
        inverted_residual_setting = [
            bneck_conf(1, 3, 1, 32, 16, 1),
            bneck_conf(6, 3, 2, 16, 24, 2),
            bneck_conf(6, 5, 2, 24, 40, 2),
            bneck_conf(6, 3, 2, 40, 80, 3),
            bneck_conf(6, 5, 1, 80, 112, 3),
            bneck_conf(6, 5, 2, 112, 192, 4),
            bneck_conf(6, 3, 1, 192, 320, 1),
        ]
        last_channel = None
    elif arch.startswith("efficientnet_v2_s"):
        inverted_residual_setting = [
            FusedMBConvConfig(1, 3, 1, 24, 24, 2),
            FusedMBConvConfig(4, 3, 2, 24, 48, 4),
            FusedMBConvConfig(4, 3, 2, 48, 64, 4),
            MBConvConfig(4, 3, 2, 64, 128, 6),
            MBConvConfig(6, 3, 1, 128, 160, 9),
            MBConvConfig(6, 3, 2, 160, 256, 15),
        ]
        last_channel = 1280
    elif arch.startswith("efficientnet_v2_m"):
        inverted_residual_setting = [
            FusedMBConvConfig(1, 3, 1, 24, 24, 3),
            FusedMBConvConfig(4, 3, 2, 24, 48, 5),
            FusedMBConvConfig(4, 3, 2, 48, 80, 5),
            MBConvConfig(4, 3, 2, 80, 160, 7),
            MBConvConfig(6, 3, 1, 160, 176, 14),
            MBConvConfig(6, 3, 2, 176, 304, 18),
            MBConvConfig(6, 3, 1, 304, 512, 5),
        ]
        last_channel = 1280
    elif arch.startswith("efficientnet_v2_l"):
        inverted_residual_setting = [
            FusedMBConvConfig(1, 3, 1, 32, 32, 4),
            FusedMBConvConfig(4, 3, 2, 32, 64, 7),
            FusedMBConvConfig(4, 3, 2, 64, 96, 7),
            MBConvConfig(4, 3, 2, 96, 192, 10),
            MBConvConfig(6, 3, 1, 192, 224, 19),
            MBConvConfig(6, 3, 2, 224, 384, 25),
            MBConvConfig(6, 3, 1, 384, 640, 7),
        ]
        last_channel = 1280
    else:
        raise ValueError(f"Unsupported model type {arch}")

    return inverted_residual_setting, last_channel


def efficientnet_b0(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b0", width_mult=1.0, depth_mult=1.0)
    return EfficientNet(inverted_residual_setting, 0.2, num_classes=num_classes,
                        last_channel=last_channel)


def efficientnet_b1(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b1", width_mult=1.0, depth_mult=1.1)
    return EfficientNet(inverted_residual_setting, 0.2, num_classes=num_classes,
                        last_channel=last_channel)


def efficientnet_b2(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b2", width_mult=1.1, depth_mult=1.2)
    return EfficientNet(inverted_residual_setting, 0.3, num_classes=num_classes,
                        last_channel=last_channel)


def efficientnet_b3(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b3", width_mult=1.2, depth_mult=1.4)
    return EfficientNet(inverted_residual_setting, 0.3, num_classes=num_classes,
                        last_channel=last_channel)


def efficientnet_b4(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b4", width_mult=1.4, depth_mult=1.8)
    return EfficientNet(inverted_residual_setting, 0.4, num_classes=num_classes,
                        last_channel=last_channel)


def efficientnet_b5(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b5", width_mult=1.6, depth_mult=2.2)
    return EfficientNet(
        inverted_residual_setting, 0.4,
        num_classes=num_classes, norm_layer=partial(nn.BatchNorm2d, eps=0.001, momentum=0.01),
        last_channel=last_channel,
    )


def efficientnet_b6(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b6", width_mult=1.8, depth_mult=2.6)
    return EfficientNet(
        inverted_residual_setting, 0.5,
        num_classes=num_classes, norm_layer=partial(nn.BatchNorm2d, eps=0.001, momentum=0.01),
        last_channel=last_channel,
    )


def efficientnet_b7(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_b7", width_mult=2.0, depth_mult=3.1)
    return EfficientNet(
        inverted_residual_setting, 0.5,
        num_classes=num_classes, norm_layer=partial(nn.BatchNorm2d, eps=0.001, momentum=0.01),
        last_channel=last_channel,
    )

def efficientnet_v2_s(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_s")
    return EfficientNet(
        inverted_residual_setting, 0.2,
        num_classes=num_classes, norm_layer=partial(nn.BatchNorm2d, eps=1e-03),
        last_channel=last_channel,
    )


def efficientnet_v2_m(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_m")
    return EfficientNet(
        inverted_residual_setting, 0.3,
        num_classes=num_classes, norm_layer=partial(nn.BatchNorm2d, eps=1e-03),
        last_channel=last_channel,
    )


def efficientnet_v2_l(num_classes):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_l")
    return EfficientNet(
        inverted_residual_setting, 0.4,
        num_classes=num_classes, norm_layer=partial(nn.BatchNorm2d, eps=1e-03),
        last_channel=last_channel,
    )


if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = efficientnet_b4(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # efficientnet_b0 Total params: 4,012,672
    # efficientnet_b1 Total params: 6,518,308
    # efficientnet_b2 Total params: 7,706,630
    # efficientnet_b3 Total params: 10,702,380
    # efficientnet_b4 Total params: 17,555,788
    # efficientnet_b5 Total params: 28,348,980
    # efficientnet_b6 Total params: 40,744,924
    # efficientnet_b7 Total params: 63,797,204
    # efficientnet_v2_s Total params: 20,182,612
    # efficientnet_v2_m Total params: 52,863,480
    # efficientnet_v2_l Total params: 117,239,396



import numpy as np
import torch
from torch import nn
from torch.nn import init



class ExternalAttention(nn.Module):

    def __init__(self, d_model,S=64):
        super().__init__()
        self.mk=nn.Linear(d_model,S,bias=False)
        self.mv=nn.Linear(S,d_model,bias=False)
        self.softmax=nn.Softmax(dim=1)
        self.init_weights()


    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, queries):
        attn=self.mk(queries) #bs,n,S
        attn=self.softmax(attn) #bs,n,S
        attn=attn/torch.sum(attn,dim=2,keepdim=True) #bs,n,S
        out=self.mv(attn) #bs,n,d_model

        return out


if __name__ == '__main__':
    input=torch.randn(50,49,512)
    ea = ExternalAttention(d_model=512,S=8)
    output=ea(input)
    print(output.shape)




import os
import torch
from tqdm import tqdm
from torch.cuda.amp import autocast
from pyzjr.dlearn import get_lr

def accuracy_all_classes(output, label):
    batch_size = label.size(0)
    _, pred = output.max(1)
    accuracy = pred.eq(label).float().sum().item() / batch_size
    return accuracy

class Classification_train_and_eval():
    """
    用于训练和评估分类模型的工具类
    """
    def __init__(self, net, train_loader, val_loader, epochs, device,
                 loss_function, optimizer, lr_scheduler, scaler):
        self.net = net
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.total_epochs = epochs
        self.device = device
        self.loss_function = loss_function
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.scaler = scaler
        self.current_valepoch = 0  # 新增一个内部变量来跟踪当前轮次
        self.current_trepoch = 0

    def evaluate(self):
        self.current_valepoch += 1
        self.net.eval()
        total_test_loss = 0
        total_correct_predictions = 0
        total_samples = 0
        total_batches = len(self.val_loader)
        with tqdm(self.val_loader, desc=f'Epoch {self.total_epochs}/{self.current_valepoch} / Test', postfix=dict, mininterval=0.3) as pbar:
            for batch_idx, data in enumerate(self.val_loader):
                img, label = data

                with torch.no_grad():
                    img = img.to(self.device)
                    label = label.to(self.device)
                    out = self.net(img)
                    test_loss = self.loss_function(out, label)
                    total_test_loss += test_loss.item()

                    _, predicted = out.max(1)
                    total_correct_predictions += predicted.eq(label).float().sum().item()
                    total_samples += label.size(0)

                pbar.set_postfix(**{'val_loss': total_test_loss / (batch_idx + 1)})
                pbar.update(1)

        self.average_test_loss = total_test_loss / total_batches
        accuracy = total_correct_predictions / total_samples

        return self.average_test_loss, accuracy


    def train_one_epoch(self):
        self.current_trepoch += 1
        self.net.train()
        total_train_loss = 0.0
        total_batches = len(self.train_loader)

        total_correct_predictions = 0
        total_samples = 0

        with tqdm(self.train_loader, desc=f'Epoch {self.total_epochs}/{self.current_trepoch} / Train', postfix=dict, mininterval=0.3) as pbar:
            for batch_idx, data in enumerate(self.train_loader):
                img, label = data
                with torch.no_grad():
                    img = img.to(self.device)
                    label = label.to(self.device)

                self.optimizer.zero_grad()
                if self.scaler is not None and torch.cuda.is_available():
                    with autocast():
                        output = self.net(img)
                        train_loss = self.loss_function(output, label)
                    self.scaler.scale(train_loss).backward()
                    self.scaler.step(self.optimizer)
                    self.scaler.update()
                else:
                    output = self.net(img)
                    train_loss = self.loss_function(output, label)
                    train_loss.backward()
                    self.optimizer.step()

                total_train_loss += train_loss.item()
                _, predicted = output.max(1)
                total_correct_predictions += predicted.eq(label).float().sum().item()
                total_samples += label.size(0)

                pbar.set_postfix(**{'total_loss': total_train_loss / (batch_idx + 1),
                                    'accuracy': total_correct_predictions / total_samples,
                                    'lr': get_lr(self.optimizer)})
                pbar.update(1)
            self.lr_scheduler.step()

        self.average_train_loss = total_train_loss / total_batches
        average_accuracy = total_correct_predictions / total_samples

        return self.average_train_loss, average_accuracy

    def save_model(self, model, save_dir='./logs', save_period=10):
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        if self.current_valepoch % save_period == 0:
            torch.save(model.state_dict(), os.path.join(save_dir,
        f"model_epoch_{self.current_valepoch}_train_{self.average_train_loss:.2f}_val_{self.average_test_loss:.2f}.pth"))

if __name__=="__main__":
    import torch.nn as nn
    import torch.optim as optim
    from torch.utils.data import DataLoader
    from torch.cuda.amp import GradScaler
    from torch.optim.lr_scheduler import StepLR
    # A simple CNN written casually, you should replace it with your own
    class SimpleCNN(nn.Module):
        def __init__(self):
            super(SimpleCNN, self).__init__()
            self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1)
            self.relu1 = nn.ReLU()
            self.pool1 = nn.MaxPool2d(kernel_size=2, stride=2)

            self.conv2 = nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1)
            self.relu2 = nn.ReLU()
            self.pool2 = nn.MaxPool2d(kernel_size=2, stride=2)

            self.fc1 = nn.Linear(128 * 8 * 8, 512)
            self.relu3 = nn.ReLU()
            self.fc2 = nn.Linear(512, 10)

    def forward(self, x):
        x = self.pool1(self.relu1(self.conv1(x)))
        x = self.pool2(self.relu2(self.conv2(x)))
        x = x.view(-1, 128 * 8 * 8)
        x = self.relu3(self.fc1(x))
        x = self.fc2(x)
        return x

    # replace with your actual dataset
    dummy_dataset = ...

    # Set up your DataLoader
    train_loader = DataLoader(dummy_dataset, batch_size=64, shuffle=True)
    val_loader = DataLoader(dummy_dataset, batch_size=64, shuffle=False)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Set up your model, loss function, optimizer, learning rate scheduler, and scaler
    model = SimpleCNN()
    model.to(device)
    loss_function = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)
    lr_scheduler = StepLR(optimizer, step_size=5, gamma=0.1)
    scaler = GradScaler()

    Epochs = 10

    classification_trainer = Classification_train_and_eval(
        net=model,
        train_loader=train_loader,
        val_loader=val_loader,
        epochs=Epochs,
        device=device,
        loss_function=loss_function,
        optimizer=optimizer,
        lr_scheduler=lr_scheduler,
        scaler=scaler
    )

    for epoch in range(Epochs):
        # Set up the Classification_train_and_eval instance
        train_loss, train_accuracy = classification_trainer.train_one_epoch()
        print(f"Epoch {epoch}/{Epochs} - Train Loss: {train_loss:.4f}, Accuracy: {train_accuracy:.4f}")

        val_loss, val_accuracy = classification_trainer.evaluate()
        print(f"Epoch {epoch}/{Epochs} - Validation Loss: {val_loss:.4f}, Accuracy: {val_accuracy:.4f}")

        classification_trainer.save_model(model)


"""
Original paper addresshttps: <https://arxiv.org/pdf/2108.01072.pdf>
"""
import torch
from torch import nn

def spatial_shift1(x):
    b, w, h, c = x.size()
    x[:, 1:, :, :c//4] = x[:, :w-1, :, :c//4]
    x[:, :w-1,:, c//4:c//2] = x[:, 1:, :, c//4:c//2]
    x[:, :,1:, c//2:c*3//4] = x[:, :, :h-1, c//2:c*3//4]
    x[:, :, :h-1, 3*c//4:] = x[:, :, 1:, 3*c//4:]
    return x


def spatial_shift2(x):
    b,w,h,c = x.size()
    x[:, :, 1:, :c//4] = x[:, :, :h-1, :c//4]
    x[:, :, :h-1, c//4:c//2] = x[:, :, 1:, c//4:c//2]
    x[:, 1:, :, c//2:c*3//4] = x[:, :w-1, :, c//2:c*3//4]
    x[:, :w-1, :, 3*c//4:] = x[:, 1:, :, 3*c//4:]
    return x


class SplitAttention(nn.Module):
    def __init__(self,channel=512,k=3):
        super().__init__()
        self.channel=channel
        self.k = k
        self.mlp1 = nn.Linear(channel, channel, bias=False)
        self.gelu = nn.GELU()
        self.mlp2 = nn.Linear(channel, channel*k, bias=False)
        self.softmax = nn.Softmax(1)

    def forward(self,x_all):
        b, k, h, w, c = x_all.shape
        x_all = x_all.reshape(b, k, -1, c)    #bs,k,n,c
        a = torch.sum(torch.sum(x_all, 1), 1)    #bs,c
        hat_a = self.mlp2(self.gelu(self.mlp1(a)))    #bs,kc
        hat_a = hat_a.reshape(b, self.k, c)    #bs,k,c
        bar_a = self.softmax(hat_a)   #bs,k,c
        attention = bar_a.unsqueeze(-2)   # #bs,k,1,c
        out = attention * x_all    # #bs,k,n,c
        out = torch.sum(out, 1).reshape(b,h,w,c)
        return out


class S2Attention(nn.Module):

    def __init__(self, channels=512 ):
        super().__init__()
        self.mlp1 = nn.Linear(channels,channels*3)
        self.mlp2 = nn.Linear(channels,channels)
        self.split_attention = SplitAttention()

    def forward(self, x):
        b, c, w, h = x.size()
        x = x.permute(0,2,3,1)
        x = self.mlp1(x)
        x1 = spatial_shift1(x[:, :, :, :c])
        x2 = spatial_shift2(x[:, :, :, c:c*2])
        x3 = x[:, :, :, c*2:]
        x_all = torch.stack([x1, x2, x3], 1)
        a = self.split_attention(x_all)
        x = self.mlp2(a)
        x = x.permute(0,3,1,2)
        return x

if __name__ == '__main__':
    input=torch.randn(50,512,7,7)
    s2att = S2Attention(channels=512)
    output=s2att(input)
    print(output.shape)



import numpy as np
import torch
from torch import nn
from torch.nn import init
from torch.nn.parameter import Parameter


class ShuffleAttention(nn.Module):

    def __init__(self, channel=512,reduction=16,G=8):
        super().__init__()
        self.G=G
        self.channel=channel
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.gn = nn.GroupNorm(channel // (2 * G), channel // (2 * G))
        self.cweight = Parameter(torch.zeros(1, channel // (2 * G), 1, 1))
        self.cbias = Parameter(torch.ones(1, channel // (2 * G), 1, 1))
        self.sweight = Parameter(torch.zeros(1, channel // (2 * G), 1, 1))
        self.sbias = Parameter(torch.ones(1, channel // (2 * G), 1, 1))
        self.sigmoid=nn.Sigmoid()


    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)


    @staticmethod
    def channel_shuffle(x, groups):
        b, c, h, w = x.shape
        x = x.reshape(b, groups, -1, h, w)
        x = x.permute(0, 2, 1, 3, 4)

        # flatten
        x = x.reshape(b, -1, h, w)

        return x

    def forward(self, x):
        b, c, h, w = x.size()
        #group into subfeatures
        x=x.view(b*self.G,-1,h,w) #bs*G,c//G,h,w

        #channel_split
        x_0,x_1=x.chunk(2,dim=1) #bs*G,c//(2*G),h,w

        #channel attention
        x_channel=self.avg_pool(x_0) #bs*G,c//(2*G),1,1
        x_channel=self.cweight*x_channel+self.cbias #bs*G,c//(2*G),1,1
        x_channel=x_0*self.sigmoid(x_channel)

        #spatial attention
        x_spatial=self.gn(x_1) #bs*G,c//(2*G),h,w
        x_spatial=self.sweight*x_spatial+self.sbias #bs*G,c//(2*G),h,w
        x_spatial=x_1*self.sigmoid(x_spatial) #bs*G,c//(2*G),h,w

        # concatenate along channel axis
        out=torch.cat([x_channel,x_spatial],dim=1)  #bs*G,c//G,h,w
        out=out.contiguous().view(b,-1,h,w)

        # channel shuffle
        out = self.channel_shuffle(out, 2)
        return out


if __name__ == '__main__':
    input=torch.randn(50,512,7,7)
    se = ShuffleAttention(channel=512,G=8)
    output=se(input)
    print(output.shape)

